import yaml

from jenkins_lockable_resources.config import Config, ConfigLoader


def test_config_update_merge_dicts():
    c = Config({"a": {"x": 1}, "b": [1, 2]}, policy="overwrite")
    c.update({"a": {"y": 2}, "b": [3], "c": 4})
    assert c["a"]["x"] == 1
    assert c["a"]["y"] == 2
    assert c["b"] == [3]
    assert c["c"] == 4


def test_config_update_type_mismatch_strict():
    c = Config({"a": 1}, policy="strict")
    try:
        c.update({"a": {"x": 1}})
        assert False, "should have raised TypeError"
    except TypeError:
        pass


def test_config_loader_reads_files(tmp_path, monkeypatch):
    # create two temp yaml files
    f1 = tmp_path / "one.yml"
    f2 = tmp_path / "two.yml"
    f1.write_text(yaml.safe_dump({"cmd": {"param1": "v1"}}))
    f2.write_text(yaml.safe_dump({"cmd": {"param2": "v2"}}))

    loader = ConfigLoader([str(f1), str(f2)])
    conf = loader.conf
    assert "cmd" in conf
    assert conf["cmd"]["param1"] == "v1"
    assert conf["cmd"]["param2"] == "v2"

    # click_settings should return a dict containing default_map
    cs = loader.click_settings
    assert isinstance(cs, dict)
    assert "default_map" in cs



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#

"""
Jenkins Lockable Resources plugin API

This library aims to query Jenkins Lockable Resources plugin to retieve information or control resources.
"""

import json
import re

from jenkinsapi.jenkinsbase import JenkinsBase


PLUGIN_NAME = "lockable-resources"


def check_request(f):
    def wrapper(*args, **kwargs):
        response = f(*args, **kwargs)
        if not response.ok:
            response.raise_for_status()

    return wrapper


class Resource:
    def __init__(
        self,
        jenkins_obj,
        name,
        *,
        is_locked=False,
        is_reserved=False,
        reserved_by=None,
        labels=None,
        is_queued=False,
        description=None,
        timestamp=None,
    ):
        self.jenkins = jenkins_obj.get_jenkins_obj()
        self.name = name
        self.is_locked = is_locked
        self.is_reserved = is_reserved
        self.reserved_by = reserved_by
        self.labels = labels
        self.baseurl = jenkins_obj.baseurl
        self.is_queued = is_queued
        self.description = description
        self.timestamp = timestamp

    @property
    def is_free(self):
        """
        Check if resource is free
        """
        return not self.is_reserved and not self.is_locked

    @property
    def is_owned(self):
        """
        Check if resource is owned by current user
        """
        return self.reserved_by == self.jenkins.username

    @check_request
    def reserve(self):
        """
        Reserve the resource
        """
        url = f"{self.baseurl}/reserve?resource={self.name}"
        return self.jenkins.requester.post_url(url)

    @check_request
    def unreserve(self):
        """
        Unreserve the resource
        """
        url = f"{self.baseurl}/unreserve?resource={self.name}"
        return self.jenkins.requester.post_url(url)

    def __str__(self):
        return self.name

    def __repr__(self):
        return str(
            {
                "name": self.name,
                "is_locked": self.is_locked,
                "is_reserved": self.is_reserved,
                "reserved_by": self.reserved_by,
                "labels": self.labels,
                "is_queued": self.is_queued,
                "description": self.description,
                "timestamp": self.timestamp,
            }
        )


GROOVY_LIST_SCRIPT = """
import org.jenkins.plugins.lockableresources.LockableResourcesManager
import groovy.json.JsonBuilder

def manager = LockableResourcesManager.get()
def resources = manager.getResources()

def resourceData = resources.collect { r ->
    return [
        name: r.getName(),
        is_locked: r.isLocked(),
        is_reserved: r.isReserved(),
        is_queued: r.isQueued(),
        reserved_by: r.getReservedBy() ?: "",
        labels: r.getLabelsAsList(),
        timestamp: r.getReservedTimestamp() ?: 0,
        description: r.getDescription() ?: "",
    ]
}

new JsonBuilder(resourceData)
"""


class LockableResources(JenkinsBase):
    """
    Class to hold information on lockable resources
    """

    def __init__(
        self,
        jenkins_obj,
        baseurl=None,
        poll=True,
        res_filter=None,
    ):
        """
        Init a lockable resource object

        Args:
            jenkins_obj (Jenkins): ref to the jenkins obj
            baseurl (str): basic url for querying information on a node
                If url is not set - object will construct it itself. This is
                useful when node is being created and not exists in Jenkins yet
            poll (bool): set to False if node does not exist or automatic
                refresh from Jenkins is not required. Default is True.
                If baseurl parameter is set to None - poll parameter will be
                set to False
            res_filter (str): Regex expression to filter resources (default '.*')
        """

        self.jenkins = jenkins_obj
        self.plugin = self.get_plugin()
        if not baseurl:
            poll = False
            baseurl = f"{self.jenkins.baseurl}/{PLUGIN_NAME}/"

        self._filter = re.compile(res_filter) if res_filter else re.compile(".*")
        self._cache_data = {}
        self._data = None

        JenkinsBase.__init__(self, baseurl, poll=poll)

    def get_plugin(self):
        # Check plugin availablility and version
        plugins = self.get_jenkins_obj().plugins
        return plugins[PLUGIN_NAME]

    def get_jenkins_obj(self):
        return self.jenkins

    def get_data(self, url, params=None, tree=None):
        """
        Retrieve data of lockable-resources
        """
        requester = self.get_jenkins_obj().requester
        response = requester.post_url(
            url=f"{self.jenkins.baseurl}/scriptText",
            data={"script": GROOVY_LIST_SCRIPT},
            verify=True,
        )
        response.raise_for_status()

        res = response.text.replace("Result: ", "").strip()
        return json.loads(res)

    @property
    def data(self):
        if self._data is None:
            self._data = self.poll()
        return self._data

    def list_resources(self) -> list:
        """
        Lists resources names
        """
        return list(self.keys())

    def get_resources(self) -> list:
        """
        Get resources
        """
        return list(self.values())

    def is_reserved(self, name) -> bool:
        """
        Check if a resource is reserved

        Args:
            name (str): The resource name
        """
        return self[name].is_reserved

    def is_locked(self, name) -> bool:
        """
        Check if a resource is locked

        Args:
            name (str): The resource name
        """
        return self[name].is_locked

    def is_free(self, name) -> bool:
        """
        Check if resource is free
        """
        return self[name].is_free

    def get_owner(self, name) -> str:
        """
        Get resource owner

        Args:
            name (str): The resource name
        """
        return self[name].reserved_by

    def get_owned_resources(self, user=None) -> list:
        """
        Get resources owned by user

        Args:
            user (str): The owner name. Default to the current jenkins user.
        """
        if user is None:
            user = self.jenkins.username

        return [
            res
            for res in self.values(reserved_by=user)
            if res.is_reserved or res.is_locked
        ]

    def get_free_resources(self) -> list:
        """
        Find a free resource
        """
        return list(self.values(is_locked=False, is_reserved=False))

    def reserve(self, name):
        """
        Reserve a resource
        """
        self[name].reserve()

    def unreserve(self, name):
        """
        Unreserve a resource
        """
        self[name].unreserve()

    def items(self):
        """
        Iterate over the names & objects for all resources
        """
        for resource in self.values():
            yield resource.name, resource

    def keys(self):
        """
        Iterate over the names of all available resources
        """
        for row in self.data:
            yield row["name"]

    def values(
        self,
        *,
        name=None,
        labels=None,
        is_locked=None,
        is_reserved=None,
        reserved_by=None,
    ):
        """
        Iterate over all available resources

        Args:
            name (str): a regex match string for resource name
            labels (str): a regex match string for resource labels
            locked (bool): if locked
            reserved (bool): if reserved
        """
        data = self.data
        if name is not None:
            if isinstance(name, str):
                name = re.compile(name)
            data = filter(lambda x: name.match(x["name"]), data)
        if labels is not None:
            if isinstance(labels, str):
                labels = re.compile(labels)
            data = filter(lambda x: labels.match(",".join(x["labels"])), data)
        if is_locked is not None:
            data = filter(lambda x: x["is_locked"] == is_locked, data)
        if is_reserved is not None:
            data = filter(lambda x: x["is_reserved"] == is_reserved, data)
        if reserved_by is not None:
            data = filter(lambda x: x["reserved_by"] == reserved_by, data)

        for attrs in data:
            yield self._make_resource(attrs)

    def _make_resource(self, attrs):
        return Resource(self, **attrs)

    def __contains__(self, resource):
        """
        True if resource exists in Jenkins
        """
        return resource in self.keys()

    def __len__(self):
        return len(self.data)

    def __getitem__(self, key):
        # Regex key search
        pattern = re.compile(key)
        for attrs in self.data:
            if pattern.match(attrs["name"]):
                return self._make_resource(attrs)
        raise KeyError(key)

    def __str__(self):
        return self.baseurl



from unittest.mock import MagicMock, patch, ANY

import click

from jenkins_lockable_resources.main import AppFactory, ClickStreamer, default_streamer


def test_appfactory_create_and_cache():
    # use a MagicMock to verify constructor is called with provided args
    ctor = MagicMock(return_value={"a": 1, "b": 2})

    f = AppFactory(ctor, 1, 2)
    # create app
    app1 = f.app
    app2 = f.app
    assert app1 is app2
    ctor.assert_called_once_with(1, 2)


def test_clickstreamer_calls_secho(mocker):
    secho = MagicMock()

    with patch("click.secho", secho):

        s = ClickStreamer()
        s.info("hello")
        s.warn("problem")
        s.error("bad")
        s.highlight("hi")

    # verify that secho was called and formatted messages exist
    secho.assert_any_call("hello", fg=ANY)
    secho.assert_any_call("WARN: problem", fg=ANY)
    secho.assert_any_call("ERROR: bad", fg=ANY)
    secho.assert_any_call("hi", fg=ANY)

    # default_streamer should be an instance of ClickStreamer
    assert isinstance(default_streamer, ClickStreamer)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Lockable Resources management application
"""
import logging
import functools

import requests.exceptions
from jenkinsapi.jenkins import Jenkins

from .model import LockableResources

INFO_STATE_COLORS = {"FREE": "green", "RESERVED": "yellow", "LOCKED": "red"}


def api_method(func):
    log = logging.getLogger(func.__name__)

    @functools.wraps(func)
    def wrapped(api, *args, **kwargs):
        try:
            return func(api, *args, **kwargs)
        except requests.exceptions.ConnectionError as e:
            api.output.error(
                f"Failed to connect to {e.request.url}. Check your connection and try again."
            )
            log.error(e)
        except Exception as e:  # pylint: disable=broad-except
            api.output.error(f"{e}")
            log.exception(e)
        return None

    return wrapped


class NoResourceError(Exception):
    """
    No resource found exception
    """


class Application:
    """
    Base Application context class for accessing.
    """

    def __init__(self, output, obj, interactive=False):
        """
        Instantiate the application

        Args:
            output: The outputer object
            obj: A backend object
            interactive: Set interactive to let prompt for user input
        """
        self.output = output
        self.interactive = interactive
        self.obj = obj


class LockableResourceApp(Application):
    """
    Application context class for accessing lockable resources in jenkins.
    """

    @staticmethod
    def from_default(
        *,
        output,
        jenkins_url,
        jenkins_user,
        jenkins_token,
        filter_expr=None,
        interactive=False,
    ):
        """
        Default app creator using dependency injections
        - Instantiate Jenkins object
        - Instantiate LockableResources
        - Instantiate LockableResourceApp

        Args:
            output: The outputer object.
            jenkins_url: The url to jenkins instance
            jenkins_user: The user to authenticate to jenkins
            jenkins_token: The user token to use for authentication
            filter_expr: A filter expression to include only matching patterns of resources
            interactive: Set interactive to let prompt for user input

        Return:
            LockableResourceApp object
        """
        # Instantiate jenkins object
        jenkins = Jenkins(jenkins_url, jenkins_user, jenkins_token, lazy=True)
        mgr = LockableResources(jenkins, res_filter=filter_expr)

        return LockableResourceApp(output, mgr, interactive)

    @api_method
    def reserve(self, name=None, labels=None, force=None):
        """
        Reserve a resource

        Args:
            name: Resource name to match (regex string)
            labels: Resource labels to match (regex string)
            force: Force reserve if already one or more resources owned
        """
        # Issue a warning if you already have one resource owned
        owned = self.obj.get_owned_resources()
        if owned:
            self.output.warn("You already have one resource owned.")
            if force is None:
                if not self.output.confirm("Force reserving?"):
                    return
            elif not force:
                return

        # Find free resources in list
        reslist = self._get_matching_free_resources(name, labels)
        if reslist and not name:
            reslist = reslist[:1]

        if not reslist:
            self.output.warn("Sorry, no free resources at the moment. Try again later.")
            return

        for res in reslist:
            res.reserve()
            self.output.info(f"Reserved {res.name}")

    @api_method
    def unreserve(self, name=None, labels=None):
        """
        Unreserve a resource

        Args:
            name: Resource name to match (regex string)
            labels: Resource labels to match (regex string)
        """
        # Name provided: Find matching resources
        reslist = self._get_matching_resources(name, labels)

        # Find owned resources in list
        reslist = [r for r in reslist if r.is_reserved]

        if not reslist:
            self.output.warn("No resources to release")
            return

        for res in reslist:
            res.unreserve()
            self.output.info(f"Unreserved {res.name}")

    @api_method
    def list(self, name=None, labels=None, state=None, short_name=False):
        """
        List resources

        Args:
            name: Resource name to match (regex string)
            labels: Resource labels to match (regex string)
            state: Match state given (case insensitive)
            short_name: Output to the short name version instead of full hostname
        """
        if state is None:
            is_locked = None
            is_reserved = None
        else:
            is_reserved = state.lower() == "reserved"
            is_locked = state.lower() == "locked"

        for res in self.obj.values(
            name, labels, is_locked=is_locked, is_reserved=is_reserved
        ):
            name = res.name
            if short_name:
                name = name.split(".")[0]
            self.output.info(name)

    @api_method
    def info(self, name=None, labels=None, state=None, reserved_by=None):
        """
        Show info of resources

        Args:
            name: Resource name to match (regex string)
            labels: Resource labels to match (regex string)
            locked: Match state given (case insensitive)
            reserved: Match state given (case insensitive)
            reserved_by: Match owner of resource (string)
        """
        if state is None:
            is_locked = None
            is_reserved = None
        else:
            is_reserved = state.lower() == "reserved"
            is_locked = state.lower() == "locked"

        for res in self.obj.values(
            name=name,
            labels=labels,
            is_locked=is_locked,
            is_reserved=is_reserved,
            reserved_by=reserved_by,
        ):
            state = (
                "LOCKED" if res.is_locked else "RESERVED" if res.is_reserved else "FREE"
            )
            self.output.info(f"{res.name}", nl=False)
            if res.labels:
                labels = ",".join(res.labels)
                self.output.info(f" [{labels}]", nl=False)
            self.output.info(": ", nl=False)
            self.output.info(f"{state}", fg=INFO_STATE_COLORS[state], nl=False)
            if not res.is_free:
                self.output.info(f" by {res.reserved_by}", nl=False)
            self.output.info("")

    @api_method
    def owned(
        self,
        user=None,
        short_name=False,
        count=None,
        index=None,
    ):
        """
        List owned resources

        Args:
            user: Owner of resource
            short_name: Output to the short name version instead of full hostname
            count: the max number of owned resources to return
            index: the owned resource list position to select
        """
        owned = self.obj.get_owned_resources(user)
        count = count or len(owned)
        if index is None:
            index = 0
        else:
            count = 1
        end = index + count

        for res in owned[index:end]:
            name = res.name
            if short_name:
                name = name.split(".")[0]
            self.output.info(name)

    def _get_matching_resources(self, name=r".*", labels=None):
        # Find matchingn resources
        reslist = list(self.obj.values(name, labels))
        if not reslist:
            exprs = []
            if name:
                exprs.append(f"name~=/{name}/")
            if labels:
                exprs.append(f"labels~=/{labels}/")
            raise NoResourceError("No resources matching " + " or ".join(exprs))
        return reslist

    def _get_matching_free_resources(self, name=r".*", labels=None):
        # Find matchingn resources
        reslist = self._get_matching_resources(name, labels)
        # Find free resources in list
        return [r for r in reslist if r.is_free]



from unittest.mock import MagicMock

import requests

from jenkins_lockable_resources.app import api_method, LockableResourceApp


def _make_conn_error():
    return requests.exceptions.ConnectionError(request=requests.Request(url="URL"))


def test_api_method_connection_error():
    out = MagicMock()

    class C:
        def __init__(self, output):
            self.output = output

        @api_method
        def f(self):
            raise _make_conn_error()

    c = C(out)
    assert c.f() is None
    out.error.assert_called_once_with(
        "Failed to connect to URL. Check your connection and try again."
    )


def test_api_method_generic_exception():
    out = MagicMock()

    class C:
        def __init__(self, output):
            self.output = output

        @api_method
        def f(self):
            raise Exception("boom")

    c = C(out)
    assert c.f() is None
    out.error.assert_called_once_with("boom")


def resource_sampler(
    name, is_free=True, is_locked=False, is_reserved=False, reserved_by="", labels=None
):
    res = MagicMock()
    res.name = name
    res.is_free = is_free
    res.is_locked = is_locked
    res.is_reserved = is_reserved
    res.reserved_by = reserved_by
    res.labels = labels or []
    return res


def resource_manager_sampler(resources, owned_resources=None):
    class Mgr:
        def values(self, *a, **k):
            return iter(resources)

        def get_owned_resources(self, user=None):
            return owned_resources

    mgr = Mgr()

    return mgr


def test_reserve_success_single_free_resource():
    out = MagicMock()
    resource = resource_sampler("res1", is_free=True, is_reserved=False, reserved_by="")
    msgr = resource_manager_sampler([resource])

    app = LockableResourceApp(out, msgr)
    app.reserve()

    resource.reserve.assert_called_once()
    out.info.assert_any_call(f"Reserved {resource.name}")


def test_unreserve_calls_unreserve_on_reserved_resources():
    out = MagicMock()

    resource = resource_sampler(
        "res2",
        is_free=False,
        is_reserved=True,
        reserved_by="user1",
        labels=["l1"],
    )
    mgr = resource_manager_sampler([resource])

    app = LockableResourceApp(out, mgr)
    app.unreserve()

    resource.unreserve.assert_called_once()
    out.info.assert_any_call(f"Unreserved {resource.name}")


def test_info_outputs_state_and_labels():
    out = MagicMock()

    resource = resource_sampler(
        "host.example.com",
        is_free=False,
        is_locked=False,
        is_reserved=True,
        reserved_by="bob",
        labels=["l1", "l2"],
    )
    mgr = resource_manager_sampler([resource])

    app = LockableResourceApp(out, mgr)
    app.info()

    # resource name printed without newline first
    out.info.assert_any_call(resource.name, nl=False)
    # state printed with color
    out.info.assert_any_call(
        "",
    )
